#include "cpu_gpgpu.h"

template <typename Type>
void CPUgpgpu<Type>::LogInfo(void)  // Borrowed from www.koders.com (under GNU)
{
#ifdef INFO_LOG
	using namespace std;

	string ProcString = "Unknown";

	LONG result;
	char value[1024];
	char buf[1024];
	DWORD valueSize;
	HKEY hKey;

	result = ::RegOpenKeyEx (HKEY_LOCAL_MACHINE, "Hardware\\Description\\System\\CentralProcessor\\0", 0, KEY_QUERY_VALUE, &hKey);
	if (result == ERROR_SUCCESS)
	{
		ProcString = " ";
		// get processor name
		valueSize = 1024;
		result = ::RegQueryValueEx (hKey, _T("ProcessorNameString"), NULL, NULL, (LPBYTE)&value, &valueSize);
		if (result == ERROR_SUCCESS)
			ProcString += value;
		else
			ProcString += "UnknownProc";
		ProcString += " ( ";
		
		// get processor frequence
		result = ::RegQueryValueEx (hKey, _T("~MHz"), NULL, NULL, (LPBYTE)&value, &valueSize);
		if (result == ERROR_SUCCESS)
		{
//			ProcString += _itoa (*(int *)value, value, 10);
			_itoa_s( *(int *)value, buf, 1024, 10 );
			ProcString += buf;
			ProcString += "MHz";
		}
		else
			ProcString += "UnknownFreq";
		ProcString += " ) ";
	}
	RegCloseKey (hKey); // Make sure to close the reg key

	ProcString += "\n\n";
	OutputNote(*ProcString.c_str());
#endif
}

template <typename Type>
void CPUgpgpu<Type>::ComputeMulMat(int mul)
{
	int idx;
	for (int i=0; i<height; i++) // submat
		for (int j=0; j<width; j++)
		{
			idx=i*width+j;
			inMatrix[idx] = inMatrix[idx]*mul;
		}
}

template <typename Type>
time_t CPUgpgpu<Type>::Compute(void)
{
//	int size;
	int tpj;
	time_t timeSum, timeStart, timeEnd;
	Type pivot;
//	width<height ? size=width : size=height;
	Type* tmpMatrix=new Type[height];
	for (int i=0; i<height; i++)
		tmpMatrix[i]=0;

	timeStart=GetTickCount();
//	int t=(height-1)/2;
	for (int t=0; t<height-1; t++) 
	{
		#ifdef DEBUG_LOG
			OutputData();
		#endif
		pivot = inMatrix[t+t*width];

//		for (int i=(t+1); i<height; i++)  // row
		for (int i=0; i<height; i++)  // row
//			if (i>=(t+1))
				tmpMatrix[i] = (Type) inMatrix[i*width+t] / (Type) pivot;
//				tmpMatrix[i]=inMatrix[i*width+t];
//			else
//				tmpMatrix[i]=-255;
//			tmpMatrix[i]=pivot;
//		for (int i=(t+1); i<height; i++) // submat
//			for (int j=t; j<width; j++)

		#ifdef DEBUG_LOG
			OutputData(tmpMatrix,1,height);
		#endif

		tpj=t+1;
//		for (int i=0; i<height; i++) // submat
//			for (int j=0; j<width; j++)
//				if (i>=tpj && j>=t)
		for (int i=tpj; i<height; i++) // submat
			for (int j=t; j<width; j++)
					inMatrix[i*width+j] = (Type) inMatrix[i*width+j] - (Type) tmpMatrix[i] * (Type) inMatrix[t*width+j];
//					inMatrix[i*width+j] = tmpMatrix[i] * inMatrix[t*width+j]; // DEBUG
//					inMatrix[i*width+j] = inMatrix[t*width+j];
//				else
//inMatrix[i*width+j] = -255;
//				inMatrix[i*width+j]=tmpMatrix[i];

	}
	timeEnd=GetTickCount();
	delete[] tmpMatrix;
	timeSum = ( timeEnd-timeStart );
	return timeSum;
}

template <typename Type>
void CPUgpgpu<Type>::ComputeTest( double a, double b )
{
	inMatrix[(height-1)*width+width-4] = (Type) a;
	inMatrix[(height-1)*width+width-3] = (Type) b;
	inMatrix[(height-1)*width+width-2] = inMatrix[(height-1)*width+width-4] / inMatrix[(height-1)*width+width-3]; // Compute already converted
	inMatrix[(height-1)*width+width-1] = (Type) ( a / b ); // Compute in double than convert
}

template <typename Type>
void CPUgpgpu<Type>::ReloadData() // Reinit inMatrix
{
    for (int j=0; j<height; j++) 
		for (int i=0; i<width; i++) 
			inMatrix[j*width+i] = relMatrix[j*width+i];
}

template <typename Type>
void CPUgpgpu<Type>::InputData( char * fileName, bool fake ) // Read the input
{
	FILE * fp;
	float input;

	if (fopen_s (&fp, fileName, "r")!=0)
		Help();

	fscanf_s(fp,"%d %d\n",&width,&height);

	inMatrix=new Type[width*height];
	relMatrix=new Type[width*height];

	if (fake)
		return;

    for (int j=0; j<height; j++) 
	{
		for (int i=0; i<width; i++) 
		{
			fscanf_s(fp,"%f ", &input);
			relMatrix[j*width+i] = (Type) input/INPUT_DIV;
			inMatrix[j*width+i] = (Type) relMatrix[j*width+i];
		}
		fscanf_s(fp,"\n");
	}
	fclose(fp);
}

template <typename Type>
void CPUgpgpu<Type>::OutputData( char * fileName, char * precision )
{
	OutputData( NULL, -1, -1, fileName, precision );
}

template <typename Type>
void CPUgpgpu<Type>::OutputData(Type * src, int w, int h, char * fileName, char * precision ) // Calculation result in file output 
{
	FILE * fp;
	Type * fColor;
	DWORD fpitch;
	static int dbgCnt=0;

	if ( fileName )
		fopen_s (&fp, fileName, "a");		
	else
		fopen_s (&fp, "output.txt", "a");

	if (w==-1 && h==-1)
	{
		w=width;
		h=height;
	}

	if ( !precision )
		precision=LOG_TYPE;

	fColor=inMatrix;
	fpitch=w;

	if (src != NULL)
	{
		fprintf(fp,"%d Debug output:\n",dbgCnt);
		dbgCnt++;
		fColor = src;
	}

	// File output 
	int fj=0;
	int fi=0;
	#ifndef LOG_WHOLE_TABLE
		if (h>1)
			fj=h-1;
		if (w>4)
			fi=w-4;
	#endif
	fColor+=fpitch*fj;
	for (int j=fj; j<h; j++) {
		for (int i=fi; i<w; i++) {
			fprintf(fp, precision, fColor[i]);
			if (i<w-1)
				fprintf(fp,", ");
		}
		fprintf(fp,"\n");
		fColor+=fpitch;
//		fColor+=fpitch/sizeof(CUTEX);
	}

//	fprintf(fp,"\n");
	fclose(fp);
}

template <typename Type>
CPUgpgpu<Type>::CPUgpgpu() // Constructor
{
}

template <typename Type>
CPUgpgpu<Type>::~CPUgpgpu() // Destructor
{
	delete[] inMatrix;
	delete[] relMatrix;
}

template class CPUgpgpu <float>; // explicit instantiation
template class CPUgpgpu <double>;
