#include "directx10_gpgpu.h"

void DX10gpgpu::InitDirect3D(HWND hWnd) // DirectX Graphics (Direct3D) initialization
{
#ifdef COMPILE_DX10
	UINT createDeviceFlags = 0;
#ifdef DEBUG_DX
    createDeviceFlags |= D3D10_CREATE_DEVICE_DEBUG;
#endif

	lastRet=D3D10CreateDevice (NULL , D3D10_DRIVER_TYPE_HARDWARE, NULL , createDeviceFlags, D3D10_SDK_VERSION, &pD3DDev); // We use the first adapter
    if(FAILED(lastRet))
		Error(*"D3D10CreateDevice() failed!\nError nr: %d",true,lastRet);

	// Create the effect
    DWORD dwShaderFlags = D3D10_SHADER_ENABLE_STRICTNESS; // Enables IEEE strictness
#if defined( DEBUG )
    dwShaderFlags |= D3D10_SHADER_DEBUG;
#endif

	ID3D10Blob *pErrors;
	lastRet = D3DX10CreateEffectFromFile( "gpgpugs02_DX10.fx", NULL, NULL, "fx_4_0", dwShaderFlags, 0, pD3DDev, NULL, NULL, &pEffect, &pErrors, NULL );
	if (FAILED(lastRet))
	{
		LPVOID pError = NULL;
		if( pErrors )
		{
			pError = pErrors->GetBufferPointer();
			Error(*(const char*)pError);
		}
		Error(*"Couldn't create effect from file gpgpugs02_DX10.fx\nD3DX10CreateEffectFromFile failed!\nError nr: %d",true,lastRet);
	}

	// Techniques
	pTechniqueSubMat = pEffect->GetTechniqueByName( "poctySubMat" );
	if ( FAILED_VALID(pTechniqueSubMat) )
		Error(*"GetTechniqueByName failed! Technique poctySubMat doesn't exist?");
	pTechniqueRow = pEffect->GetTechniqueByName( "poctyRow" );
	if ( FAILED_VALID(pTechniqueRow) )
		Error(*"GetTechniqueByName failed! Technique poctyRow doesn't exist?");

	// Resource
	pSRVarTexIn = pEffect->GetVariableByName("mytex")->AsShaderResource();
	pSRVarTexTemp = pEffect->GetVariableByName("mytexTemp")->AsShaderResource();
	// Variables
	pRVarPivot = pEffect->GetVariableByName("pivot")->AsScalar();
	pRVarWidth = pEffect->GetVariableByName("width")->AsScalar();
	pRVarHeight = pEffect->GetVariableByName("height")->AsScalar();
#endif
}

DX10gpgpu::DX10gpgpu() // Constructor
{
#ifdef COMPILE_DX10
	pD3DDev = NULL; // Interface pointer to video card
#endif
}

DX10gpgpu::~DX10gpgpu() // Destructor
{
#ifdef COMPILE_DX10
	if (pD3DDev)
		pD3DDev->ClearState();
	RELEASE (pD3DDev);
	RELEASE (pTexIn);
	RELEASE (pTexRel);
#endif
}

void DX10gpgpu::Render()
{
#ifdef COMPILE_DX10
	ID3D10InputLayout *	pVertexLayout;
	ID3D10Buffer * pVertexBuffer;
	ID3D10Texture2D * pTexPong = NULL;
	ID3D10Texture2D * pTexTemp = NULL;
	ID3D10Texture2D * pTexSwap = NULL;
	ID3D10ShaderResourceView * const pSRV[2] = {NULL};
	ID3D10ShaderResourceView * pSRViewPong;
	ID3D10ShaderResourceView * pSRViewIn;
	ID3D10ShaderResourceView * pSRViewTemp;
	ID3D10ShaderResourceView * pSRViewSwap;
	ID3D10RenderTargetView * pRTViewPong = NULL;
	ID3D10RenderTargetView * pRTViewTemp = NULL;
	ID3D10RenderTargetView * pRTViewIn = NULL;
	ID3D10RenderTargetView * pRTViewSwap = NULL;

	D3D10_TECHNIQUE_DESC techDescSubMat;
	D3D10_TECHNIQUE_DESC techDescRow;

	// Create a texture to use as the target render
	D3D10_TEXTURE2D_DESC descPong, descTemp;
	ZeroMemory( &descPong, sizeof(descPong) );
	descPong.Width = width;
	descPong.Height = height;
	descPong.MipLevels = 1;
	descPong.ArraySize = 1;
	descPong.Format = DXGI_FORMAT_R32_FLOAT;
	descPong.SampleDesc.Count = 1;
	descPong.Usage = D3D10_USAGE_DEFAULT;
	descPong.BindFlags = D3D10_BIND_RENDER_TARGET | D3D10_BIND_SHADER_RESOURCE;

	descTemp = descPong;
	descTemp.Width = 1;

	lastRet=pD3DDev->CreateTexture2D(&descPong,NULL,&pTexPong); 
    if(FAILED(lastRet))
		Error(*"dx10 CreateTexture2D pongtex failed!\nError nr: %d",true,lastRet);

	lastRet=pD3DDev->CreateTexture2D(&descTemp,NULL,&pTexTemp); 
    if(FAILED(lastRet))
		Error(*"dx10 CreateTexture2D temptex failed!\nError nr: %d",true,lastRet);

	// Create the RenderTargetView
	D3D10_RENDER_TARGET_VIEW_DESC descRT;
	descRT.Format = descPong.Format;
	descRT.ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2D;
	descRT.Texture2DArray.MipSlice = 0;

	pD3DDev->CreateRenderTargetView( pTexPong, &descRT, &pRTViewPong );
	pD3DDev->CreateRenderTargetView( pTexTemp, &descRT, &pRTViewTemp );
	pD3DDev->CreateRenderTargetView( pTexIn, &descRT, &pRTViewIn );

	// Create a ShaderResourceView that we need to be used by the shader.
	D3D10_SHADER_RESOURCE_VIEW_DESC srDesc;
	srDesc.Format = descPong.Format;
	srDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
	srDesc.Texture2D.MostDetailedMip = 0;
	srDesc.Texture2D.MipLevels = 1;

	pD3DDev->CreateShaderResourceView( pTexPong, &srDesc, &pSRViewPong ); 
	pD3DDev->CreateShaderResourceView( pTexIn, &srDesc, &pSRViewIn ); 
	pD3DDev->CreateShaderResourceView( pTexTemp, &srDesc, &pSRViewTemp ); 

	// Define the input layout
	const D3D10_INPUT_ELEMENT_DESC defaultLayout[] =
	{
		{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
	};
	UINT numDefaultElements = sizeof(defaultLayout)/sizeof(D3D10_INPUT_ELEMENT_DESC); // = 2

	// Create and set the vertex input layout
	D3D10_PASS_DESC passDesc;
	pTechniqueSubMat->GetPassByIndex(0)->GetDesc(&passDesc);
	if (FAILED(pD3DDev->CreateInputLayout(defaultLayout, numDefaultElements, passDesc.pIAInputSignature, passDesc.IAInputSignatureSize, &pVertexLayout)))
		Error(*"CreateInputLayout failed");
	pD3DDev->IASetInputLayout(pVertexLayout);

	// Create vertex buffer
	CUVERTEX3 myVertices[] = // Define texcoords: x, y, z position points +  2 texture coord
    {
		// A render target is oriented with the Y axes pointing down; this requires that the Y coordinates get flipped 
		{  1.0f,  1.0f, 0.0f, 1.0f, 0.0f },
		{ -1.0f,  1.0f, 0.0f, 0.0f, 0.0f },
		{  1.0f, -1.0f, 0.0f, 1.0f, 1.0f },
		{ -1.0f, -1.0f, 0.0f, 0.0f, 1.0f },
    };

	D3D10_BUFFER_DESC bd;
	bd.Usage = D3D10_USAGE_DEFAULT;
	bd.ByteWidth = sizeof(myVertices);
	bd.BindFlags = D3D10_BIND_VERTEX_BUFFER;
	bd.CPUAccessFlags = 0;
	bd.MiscFlags = 0;  

	D3D10_SUBRESOURCE_DATA initData;
	ZeroMemory(&initData, sizeof(initData));
	initData.pSysMem = myVertices;
    
	lastRet=pD3DDev->CreateBuffer(&bd, &initData, &pVertexBuffer);
    if(FAILED(lastRet))
		Error(*"CreateBuffer failed!\nError nr: %d",true,lastRet);

	// Set vertex buffer and topology
	UINT strides;
	UINT offsets;

	strides = (UINT)sizeof(CUVERTEX3);
	offsets = 0;
	pD3DDev->IASetVertexBuffers(0, 1, &pVertexBuffer, &strides, &offsets); // Bind an array of vertex buffers to the input-assembler stage.
	pD3DDev->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);

	// Set viewport
	D3D10_VIEWPORT vp;
    vp.Width = width;
    vp.Height = height;
    vp.MinDepth = 0.0f;
    vp.MaxDepth = 1.0f;
    vp.TopLeftX = 0;
    vp.TopLeftY = 0;
    pD3DDev->RSSetViewports( 1, &vp );

	pRVarWidth->SetFloat( (float)(width) ); // Send variable to shaders
	pRVarHeight->SetFloat( (float)(height) );

	for (int t=0; t<height-1; t++) 
	{
		#ifdef DEBUG_LOG
			OutputData();
		#endif

		pRVarPivot->SetFloat((float)(t));

		// row
		pD3DDev->PSSetShaderResources(0, 2, pSRV); // Unbind textures
		pSRVarTexIn->SetResource(pSRViewIn); // Set source texture
		pD3DDev->OMSetRenderTargets(1, &pRTViewTemp, NULL); // Render to temporary pivot texture

		pTechniqueRow->GetDesc(&techDescRow);
		for(UINT p=0; p<techDescRow.Passes; ++p)
		{
			pTechniqueRow->GetPassByIndex(p)->Apply(0); // Bound our shader to a stage in the pipeline
			pD3DDev->Draw(4, 0); // Start processing the vertex data passed into the graphics pipeline
		}

		#ifdef DEBUG_LOG
			OutputData(pTexTemp,1,height);
		#endif

		// submat
		pD3DDev->PSSetShaderResources(0, 2, pSRV); // Unbind textures
		pSRVarTexTemp->SetResource(pSRViewTemp); // Set source texture
		pSRVarTexIn->SetResource(pSRViewIn); // Set source texture
		pD3DDev->OMSetRenderTargets( 1, &pRTViewPong, NULL ); // Render back to matrix (through in<->pong)

		pTechniqueSubMat->GetDesc(&techDescSubMat);
		for(UINT p=0; p<techDescSubMat.Passes; ++p)
		{
			pTechniqueSubMat->GetPassByIndex(p)->Apply(0);  // Bound our shader to a stage in the pipeline
			pD3DDev->Draw(4, 0); // Start processing the vertex data passed into the graphics pipeline
		}

		// Ping-pong (swap textures in<->pong) 
		pSRViewSwap = pSRViewPong;
		pSRViewPong = pSRViewIn;
		pSRViewIn = pSRViewSwap;		
		pRTViewSwap = pRTViewPong;
		pRTViewPong = pRTViewIn;
		pRTViewIn = pRTViewSwap;
		pTexSwap = pTexPong;
		pTexPong = pTexIn;
		pTexIn = pTexSwap;
	}

	pD3DDev->Flush();
	RELEASE(pSRViewIn); // = pSRViewSwap
	RELEASE(pSRViewPong); // zustava ref
	RELEASE(pSRViewTemp); // zustava ref
	RELEASE(pRTViewSwap);
	RELEASE(pRTViewPong);
//	RELEASE(pRTViewSwap);
//	RELEASE(pTexPong);
//	RELEASE(pTexSwap);
	RELEASE(pTexTemp);
#endif
}

void DX10gpgpu::ReloadData()
{
	if ( pTexIn==pTexRel || !pTexIn || !pTexRel)
		Error(*"pTexRel or pTexIn error",false);

 	pD3DDev->CopyResource( pTexIn, pTexRel );
}

void DX10gpgpu::InputData( char * fileName )
{
#ifdef COMPILE_DX10
	FILE * fp;
	float tmp;
//	ID3D10Texture2D * pReadTex;

	if (fopen_s (&fp, fileName, "r")!=0)
		Help();

	fscanf_s(fp,"%d %d\n",&width,&height);

	// To fill texture manually - read texture
	D3D10_TEXTURE2D_DESC desc; // Describes a 2D texture
	desc.Width = width;
	desc.Height = height;
	desc.MipLevels = desc.ArraySize = 1;
	desc.Format = DXGI_FORMAT_R32_FLOAT;
	desc.SampleDesc.Count = 1;
	desc.SampleDesc.Quality = 0;
	desc.MiscFlags = 0;
	desc.Usage = D3D10_USAGE_DYNAMIC; // CPU write, GPU read
	desc.BindFlags = D3D10_BIND_SHADER_RESOURCE;
	desc.CPUAccessFlags = D3D10_CPU_ACCESS_WRITE;

	// Copy read texture to in texture
	D3D10_TEXTURE2D_DESC descIn = desc;
	descIn.Usage = D3D10_USAGE_DEFAULT;
	descIn.BindFlags = D3D10_BIND_RENDER_TARGET | D3D10_BIND_SHADER_RESOURCE;
	descIn.CPUAccessFlags = 0;

	lastRet=pD3DDev->CreateTexture2D( &desc, NULL, &pTexRel );
    if(FAILED(lastRet))
		Error(*"dx10 CreateTexture2D readtex failed!\nError nr: %d",true,lastRet);

	lastRet=pD3DDev->CreateTexture2D( &descIn, NULL, &pTexIn );
    if(FAILED(lastRet))
		Error(*"dx10 CreateTexture2D inputtex failed!\nError nr: %d",true,lastRet);

	D3D10_MAPPED_TEXTURE2D mappedTex; // Provides access to subresource data in a 2D texture
	lastRet=pTexRel->Map( D3D10CalcSubresource(0, 0, 1), D3D10_MAP_WRITE_DISCARD, 0, &mappedTex ); // Enables the application to access the underlying memory of the texture
    if(FAILED(lastRet))
		Error(*"dx10 Map() failed!\nError nr: %d",true,lastRet);

	float* pTexels = (float*)mappedTex.pData;
	DWORD pitch = mappedTex.RowPitch/4;

    for (int j=0; j<height; j++) 
	{
		for (int i=0; i<width; i++) 
		{
			fscanf_s(fp,"%f ",&tmp); // &pTexels[i + j*width]);
			pTexels[i]=tmp/INPUT_DIV;
		}
		pTexels+=pitch;
		fscanf_s(fp,"\n");
	}
	fclose(fp);

	pTexRel->Unmap( D3D10CalcSubresource(0, 0, 1) ); // Invalidate pointer to the resource

	pD3DDev->CopyResource( pTexIn, pTexRel );

//	RELEASE (pReadTex);
#endif
}


void DX10gpgpu::OutputData( char * fileName )
{
	OutputData( NULL, -1, -1, fileName );
}

void DX10gpgpu::OutputData(ID3D10Texture2D * src, int w, int h, char * fileName )
{
#ifdef COMPILE_DX10
	FILE * fp;
	ID3D10Texture2D * pWriteTex;
	ID3D10Texture2D * pSrcTex = pTexIn;
	static int dbgCnt=0;

	if (w==-1 && h==-1)
	{
		w=width;
		h=height;
	}
	if ( !fileName )
		fopen_s (&fp, "output.txt", "a");
	else
		fopen_s (&fp, fileName, "a");

	if (src != NULL)
	{
		fprintf(fp,"%d Debug output:\n",dbgCnt);
		dbgCnt++;
		pSrcTex = src;
	}

	// To fill texture manually - write texture
	D3D10_TEXTURE2D_DESC desc; // Describes a 2D texture
	desc.Width = w;
	desc.Height = h;
	desc.MipLevels = desc.ArraySize = 1;
	desc.Format = DXGI_FORMAT_R32_FLOAT;
	desc.SampleDesc.Count = 1;
	desc.SampleDesc.Quality = 0;
	desc.MiscFlags = 0;
	desc.Usage = D3D10_USAGE_STAGING; // CPU read/write; GPU only copy
	desc.BindFlags = 0; // Staging resource could not have any bind point 
	desc.CPUAccessFlags = D3D10_CPU_ACCESS_READ;

	lastRet=pD3DDev->CreateTexture2D( &desc, NULL, &pWriteTex );
    if(FAILED(lastRet))
		Error(*"dx10 CreateTexture2D writetex failed!\nError nr: %d",true,lastRet);

	pD3DDev->CopyResource( pWriteTex, pSrcTex );

	D3D10_MAPPED_TEXTURE2D mappedTex; // Provides access to subresource data in a 2D texture
	lastRet=pWriteTex->Map( D3D10CalcSubresource(0, 0, 1), D3D10_MAP_READ, 0, &mappedTex ); // Enables the application to access the underlying memory of the texture
    if(FAILED(lastRet))
		Error(*"dx10 Map() failed!\nError nr: %d",true,lastRet);

	float* pTexels = (float*)mappedTex.pData;
	DWORD pitch = mappedTex.RowPitch/4;

	// File output 
	int fj=0;
	int fi=0;
	#ifndef LOG_WHOLE_TABLE
		if (h>1)
			fj=h-1;
		if (w>4)
			fi=w-4;
		pTexels+=pitch*fj;
	#endif
	for (int j=fj; j<h; j++) {
		for (int i=fi; i<w; i++) {
			fprintf(fp,LOG_TYPE,pTexels[i]); // pTexels[i + j*width]);
			if (i<w-1)
				fprintf(fp,", ");
		}
		pTexels+=pitch;
		fprintf(fp,"\n");
	}

	pWriteTex->Unmap( D3D10CalcSubresource(0, 0, 1) ); // Invalidate pointer to the resource

	RELEASE (pWriteTex);

//	fprintf(fp,"\n");
	fclose(fp);
#endif
}

void DX10gpgpu::LogInfo()
{
#ifdef COMPILE_DX10
#ifdef INFO_LOG
	IDXGIFactory * pFactory;
	IDXGIAdapter * pAdapter;
	DXGI_ADAPTER_DESC identifier;

	CreateDXGIFactory(__uuidof(IDXGIFactory), (void**)(&pFactory) );
	pFactory->EnumAdapters( 0, &pAdapter ); // Get pointer to first adapter (we use first in createdevice)
	pAdapter->GetDesc( &identifier );
	
	_snprintf_s( buf, sizeof(buf)/sizeof(TCHAR), MAXSTRSIZE, " %S ( %luMB )\n\n", identifier.Description, (unsigned long)identifier.DedicatedVideoMemory/1024/1024 );

	OutputNote( *buf );

#endif
#endif
}