// TODO:
//	pingpong po dorendrovani
//	pivot na nvidii ok, na notasu posunutej (pocitat v shaderech?)
//	presnost (pocitani a ulozeni v texture) - individualni podle karty
//  ?zmenseni pole - pres vertexy? ted se vzdy projizdi cela textura a submat se resi podminkama
//	zkontrolovat premazani inMatrix po RELEASE read textury

#include "directx9_gpgpu.h"

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZRHW | D3DFVF_TEX1) // Our vertex stucture define.

void DX9gpgpu::InitDirect3D(HWND hWnd) // DirectX Graphics (Direct3D) initialization
{
	int lastRet;

	pD3D = Direct3DCreate9 (D3D_SDK_VERSION); // Create the Direct3D object to get everything started.
    if(pD3D==NULL)
		Error(*"Error, couldn't initialize DirectX!?!");
    
    D3DDISPLAYMODE display_mode; // This object will allow us to set the display mode of the screen.
    if (FAILED(pD3D->GetAdapterDisplayMode (D3DADAPTER_DEFAULT, &display_mode))) // This function will get the display mode of the device and place it in DisplayMode.
		Error(*"Error setting the display mode.");

    D3DPRESENT_PARAMETERS parameters; // This will allow us to set the parameters of the screen.
    ZeroMemory (&parameters, sizeof (parameters));
    parameters.Windowed = TRUE;
    parameters.SwapEffect = D3DSWAPEFFECT_DISCARD; // Dealing with animation
//	parameters.SwapEffect = D3DSWAPEFFECT_FLIP;
    parameters.BackBufferFormat = display_mode.Format; // Render to the area of the screen.
//	parameters.BackBufferFormat = D3DFMT_UNKNOWN;

//		parameters.BackBufferCount = 1;                  // Number of back buffers.
//		parameters.EnableAutoDepthStencil = FALSE;
//		parameters.AutoDepthStencilFormat = D3DFMT_D16;
		
	if(FAILED(pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps))) // Get what this hardware is abble to do
		Error(*"GetDeviceCaps() failed!!");
		   
	if(D3DCaps.VertexProcessingCaps==0) // Test which is supported, hardware or software vertex processing.
		Error(*"Hardware vertex processing not supported by your hardware!!");

	// TODO maybe PS3, after precision test
	if(D3DCaps.VertexShaderVersion<D3DVS_VERSION(1,1) || D3DCaps.PixelShaderVersion<D3DPS_VERSION(2,0)) // Test shader version.
		Error(*"Too low shader version. Need VS 1.1 and PS 2.0 at least!!");

	lastRet=pD3D->CreateDevice (D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_FPU_PRESERVE, &parameters, &pD3DDev);

    if(FAILED(lastRet)) // Now we must create the rendering device.
	{
		Error(*"CreateDevice() failed! Window size equal to zero or no DirectX 9 installed.\nError nr: %d",true,lastRet);
		// -2147024809 = E_INVALIDARG = An invalid parameter was passed to the returning function
		// -2005530585 = D3DERR_DRIVERINTERNALERROR
		// -2005530516 = D3DERR_INVALIDCALL
	}
	if(pD3D==NULL) // One last check to be sure.
		Error(*"D3D_Device is equal to NULL!?!");
	
	pD3DDev->SetRenderState(D3DRS_LIGHTING, FALSE); // Lights off
	pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE); // Disable culling.  This will render the front and back of the triangles.
	pD3DDev->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, FALSE); // Disable antialiasing

    LPD3DXBUFFER pEB = NULL;
	if (FAILED(D3DXCreateEffectFromFile(pD3DDev, "gpgpugs02_DX9.fx", NULL, NULL, 0, NULL, &pEffect, &pEB)))
    {		
		if (pEB==NULL)
			Error(*"Error, couldn't create effect from file gpgpugs02_DX9.fx");
		else 
		{
			LPVOID compileErrors = pEB->GetBufferPointer();
			Error(*(const char*)compileErrors);
		}
	}
    RELEASE(pEB);
}

void DX9gpgpu::Render(int width, int height) // drawing function 
{
	int t=0;

	UINT totalPasses; // Used to store the total amount of passes we need for our effect.
	D3DXHANDLE hTexcoord2;
	LPDIRECT3DSURFACE9 pSwapSurf;
	LPDIRECT3DTEXTURE9 pSwapTex;

	// Correctly map from texels->pixels, offset the coordinates by -0.5f	
	float fWidth = (float)( width ) - 0.5f;
    float fHeight = (float)( height ) - 0.5f;    

    CUVERTEX myVertices[] = // Define texcoords: x, y, z, w position points +  2 texture coord
	{ 
/*		// Rectangle to big triangle optimalization
		{ -0.5f,   -0.5f,  0.0f, 1.0f, 0.0f, 0.0f},
        { fWidth*2,  -0.5f,  0.0f, 1.0f, 2.0f, 0.0f},
        { -0.5f,  fHeight*2, 0.0f, 1.0f, 0.0f, 2.0f},
*/
		// Disable optimalization to compare
		{ -0.5f,   -0.5f,  0.0f, 1.0f, 0.0f, 0.0f},
        { fWidth,  -0.5f,  0.0f, 1.0f, 1.0f, 0.0f},
        { -0.5f,  fHeight, 0.0f, 1.0f, 0.0f, 1.0f},
        { fWidth, fHeight, 0.0f, 1.0f, 1.0f, 1.0f}
    };    

	// Here we create a vertex buffer.  To draw something to the screen you
	// must put the vertices of, lets say a triangle, in this buffer and send it to SetStreamSource().    
//	if(FAILED(pD3DDev->CreateVertexBuffer(sizeof(myVertices), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &pVB, NULL)))
//		Error(*"Error, couldn't create vertex buffer");

	if (FAILED(D3DXCreateTexture (pD3DDev, 1, height, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pTempTex))) // Texture as a rendering target formation D3DFMT_R32F
		Error(*"Error, couldn't create temp texture");
	ClearTexture(pTempTex,1,height);
	pTempTex->GetSurfaceLevel (0, &pTempSurf);

	if (FAILED(D3DXCreateTexture (pD3DDev, width, height, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pPongTex))) // Texture as a rendering target formation D3DFMT_R32F
		Error(*"Error, couldn't create pong texture");
	ClearTexture(pPongTex,width,height);
	pPongTex->GetSurfaceLevel (0, &pPongSurf);

    pD3DDev->BeginScene (); // Tell Direct 3D we are now going to start drawing.
	{
//		t=(height-1)/2;
		for (t=0; t<height-1; t++) 
		{
			#ifdef DEBUG_LOG
				OutputData();
			#endif

//			hTexcoord = pEffect->GetParameterBySemantic(NULL, "COLOR2");
//			if (FAILED(pEffect->SetVector (hTexcoord,&(D3DXVECTOR4 ((float)(t+0.5)/width, (float)(t-0.5)/width, (float)(t+0.5)/height, (float)(t+1.5)/height))))) // Send variable to hlsl - pivot coordinate
//				Error(*"Error, couldn't SetVector");
			hTexcoord2 = pEffect->GetParameterBySemantic(NULL, "COLOR1");
			if (FAILED(pEffect->SetVector (hTexcoord2,&(D3DXVECTOR4 ((float)width, (float)height, (float)1, (float)(t)))))) // Send variable to hlsl - pivot coordinate
				Error(*"Error, couldn't SetVector");

			// row		
			if (FAILED(pEffect->SetTexture ("mytex", pInTex))) // Set in texture
				Error(*"Error, couldn't set in texture");

			pEffect->SetTechnique ("poctyRow");			
			if (FAILED(pD3DDev->SetRenderTarget (0, pTempSurf))) // Modification of rendering target
				Error(*"Error, couldn't SetRenderTarget");

		    pEffect->Begin (&totalPasses, 0);
			for(UINT p=0; p<totalPasses; p++)
			{         
				pEffect->BeginPass(p);
				{						
//					pD3DDev->SetStreamSource(0,pVB,0,sizeof(CUVERTEX)); // This will bind the vertex data in the buffer to the Direct3D device.
					pD3DDev->SetFVF(D3DFVF_CUSTOMVERTEX); // Set the vertex stream declaration.
					pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, myVertices, sizeof(CUVERTEX));
//					pD3DDev->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2); // This will draw everything in the buffer.
				}
				pEffect->EndPass();
			}
			pEffect->End();  // End of the effect.

			#ifdef DEBUG_LOG
				OutputData(pTempSurf,1,height);
			#endif

			// submat

			pEffect->SetTechnique ("poctySubMat");
			if (FAILED(pD3DDev->SetRenderTarget (0, pPongSurf))) // Modification of rendering target
				Error(*"Error, couldn't SetRenderTarget 2");

			if (FAILED(pEffect->SetTexture ("mytex", pInTex))) // Set in texture
				Error(*"Error, couldn't set in texture");

			if (FAILED(pEffect->SetTexture ("mytexTemp", pTempTex))) // Set temp texture
				Error(*"Error, couldn't set temp texture");		


		    pEffect->Begin (&totalPasses, 0);
			for(UINT pass=0; pass<totalPasses; pass++)
			{         
				pEffect->BeginPass(pass);
			    {						
//					pD3DDev->SetStreamSource(0,pVB,0,sizeof(CUVERTEX)); // This will bind the vertex data in the buffer to the Direct3D device.
				    pD3DDev->SetFVF(D3DFVF_CUSTOMVERTEX); // Set the vertex stream declaration.
					pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, myVertices, sizeof(CUVERTEX));
//					pD3DDev->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2); // This will draw everything in the buffer.
				}
				pEffect->EndPass ();
			}
			pEffect->End ();  // End of the effect.

			// Ping-pong (swap textures)
			pSwapSurf = pInSurf;
			pInSurf = pPongSurf;
			pPongSurf = pSwapSurf;
			pSwapTex = pInTex;
			pInTex = pPongTex;
			pPongTex = pSwapTex;

		}
	}
	pD3DDev->EndScene (); // We are done drawing for this scene.
	pD3DDev->Present (NULL, NULL, NULL, NULL); // Swap the old frame with the new one.
	RELEASE(pTempSurf);
	RELEASE(pTempTex);

	RELEASE(pPongTex);
	RELEASE(pPongSurf);
}

void DX9gpgpu::ClearTexture(LPDIRECT3DTEXTURE9 pClearTex,int width,int height)
{
	DWORD pitch;
	LPDIRECT3DTEXTURE9 pReadTex = NULL;
	LPDIRECT3DSURFACE9 pReadSurf = NULL;

	D3DXCreateTexture (pD3DDev, width, height, 1, 0, D3DFMT_R32F, D3DPOOL_SYSTEMMEM, &pReadTex);
	pReadTex->GetSurfaceLevel (0, &pReadSurf);
    D3DLOCKED_RECT rect;
	pReadTex->LockRect (0, &rect, NULL, 0);
    inMatrix = (float*) rect.pBits;
	pitch = rect.Pitch/4;

    for (int j=0; j<height; j++) 
	{
		for (int i=0; i<width; i++) 
			inMatrix[i]=0;
		inMatrix+=pitch;
	}

	if (FAILED(pD3DDev->UpdateTexture (pReadTex, pClearTex))) // copy in texture to video memory
		Error(*"Error, couldn't clear (copy) texture");

	pReadTex->UnlockRect (0);
	RELEASE (pReadTex);
	RELEASE (pReadSurf);
}

void DX9gpgpu::InputDataTest( char * fileName, double a, double b )
{
	InputData( fileName, true, a, b );
}

void DX9gpgpu::ReloadData()
{
    D3DLOCKED_RECT rect;
	pTexRel->LockRect( 0, &rect, NULL, 0 );

	if (FAILED(pD3DDev->UpdateTexture (pTexRel, pInTex))) // copy in texture to video memory
		Error(*"Error, couldn't copy texture");		
	pTexRel->UnlockRect( 0 );
}

void DX9gpgpu::InputData( char * fileName, bool first_int, double a, double b ) // Read the input (and write it to texture)
{
	FILE * fp;
	DWORD pitch;
	int w, h;
	float tmp;

	// DX9
//	LPDIRECT3DTEXTURE9 pReadTex = NULL;
	LPDIRECT3DSURFACE9 pReadSurf = NULL;

	if (fopen_s (&fp, fileName, "r")!=0)
		Help();

	fscanf_s(fp,"%d %d\n",&width,&height);

	w = width;
	h = height;

//	if (first_int)
//		height = 1;

	D3DXCreateTexture (pD3DDev, w, h, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pInTex); // Texture as a rendering target formation
	pInTex->GetSurfaceLevel (0, &pInSurf);
	D3DXCreateTexture (pD3DDev, w, h, 1, 0, D3DFMT_R32F, D3DPOOL_SYSTEMMEM, &pTexRel);
	pTexRel->GetSurfaceLevel (0, &pReadSurf);
    D3DLOCKED_RECT rect;
	pTexRel->LockRect (0, &rect, NULL, 0);
//	CUTEX* inMatrix = (CUTEX*) rect.pBits;
    inMatrix = (float*) rect.pBits;
	pitch = rect.Pitch/4;

	if ( first_int )
	{
		inMatrix+=pitch*(h-1);
		inMatrix[w-4]=(float) a;
		inMatrix[w-3]=(float) b;
	} else
	{
		for (int j=0; j<h; j++) 
		{
			for (int i=0; i<w; i++) 
			{
				fscanf_s(fp,"%f ",&tmp);
				inMatrix[i]=tmp/INPUT_DIV;
			}			
			fscanf_s(fp,"\n");
			inMatrix+=pitch;
		}
	}
	fclose(fp);

	if (FAILED(pD3DDev->UpdateTexture (pTexRel, pInTex))) // copy in texture to video memory
		Error(*"Error, couldn't copy texture");		

	pTexRel->UnlockRect (0);
//	RELEASE (pReadTex);
	RELEASE (pReadSurf);
}

void DX9gpgpu::OutputDataTest()
{
	OutputData( pInSurf, width, height, true);
}

void DX9gpgpu::OutputData( char * fileName, char * precision )
{
	OutputData(NULL, -1, -1, false, fileName, precision );
}

void DX9gpgpu::OutputData(LPDIRECT3DSURFACE9 src, int w, int h, bool first_int, char * fileName, char * precision ) // Calculation result in file output 
{
	FILE * fp;
	float* fColor;
	DWORD fpitch;
	static int dbgCnt=0;

	LPDIRECT3DTEXTURE9 pOutputTex = NULL; // Texture object to create output.
	LPDIRECT3DSURFACE9 pOutputSurf = NULL;
	LPDIRECT3DSURFACE9 pSrc = pInSurf;

	if (w==-1 && h==-1)
	{
		w=width;
		h=height;
	}

	if ( !precision )
		precision=LOG_TYPE;

	if ( fileName )
		fopen_s (&fp, fileName, "a");
	else
		fopen_s (&fp, "output.txt", "a");

	if (src != NULL && !first_int)
	{
		fprintf(fp,"%d Debug output:\n",dbgCnt);
		dbgCnt++;
		pSrc = src;
	}

	// SYSTEMMEM texture is drawn up in the territory 
	D3DXCreateTexture (pD3DDev, w, h, 1, 0, D3DFMT_R32F, D3DPOOL_SYSTEMMEM, &pOutputTex); // D3DFMT_R32F, D3DFMT_A32B32G32R32F
	pOutputTex->GetSurfaceLevel (0, &pOutputSurf);
	// SYSTEMMEM it copies in the territory 
	pD3DDev->GetRenderTargetData (pSrc, pOutputSurf);

	D3DLOCKED_RECT rect;
	pOutputTex->LockRect (0, &rect, NULL, 0); // Lock of video memory 
//	CUTEX* fColor = (CUTEX*) rect.pBits;
	fColor = (float*) rect.pBits;
	fpitch = rect.Pitch/4;

	// File output 
	if (first_int) // Only first line (in integers)
	{
		fColor+=fpitch*(h-1);
		for (int i=0; i<w; i++) 
		{
			fprintf(fp,"%d",(int)fColor[i]);
			if (i<w-1)
				fprintf(fp,", ");
		}
	} else
	{
		int fj=0;
		int fi=0;
#ifndef LOG_WHOLE_TABLE
		if (h>1)
			fj=h-1;
		if (w>4)
			fi=w-4;
#endif
		fColor+=fpitch*fj;
		for (int j=fj; j<h; j++) 
		{
			for (int i=fi; i<w; i++) 
			{
				fprintf( fp, precision, fColor[i] );
				if (i<w-1)
				fprintf(fp,", ");
			}
			fprintf(fp,"\n");
			fColor+=fpitch;
//			fColor+=fpitch/sizeof(CUTEX);
		}
	}

	pOutputTex->UnlockRect (0); // unlocking 
	RELEASE (pOutputTex);
	RELEASE (pOutputSurf);

//	fprintf(fp,"\n");
	fclose(fp);
}

DX9gpgpu::DX9gpgpu() // Constructor
{
	pD3D = NULL; // Interface pointer to of Direct3D
//	pVB = NULL; // The interface pointer to vertex buffer for hold our main object.
	pEffect = NULL;
//	pVertexDeclaration = NULL;
	pInTex = NULL; // Texture object to hold IN texture image.
	pInSurf = NULL;
	pTexRel= NULL;
	pD3DDev = NULL; // Interface pointer to video card
}

DX9gpgpu::~DX9gpgpu() // Destructor
{
	RELEASE (pEffect);
	RELEASE (pInTex);
	RELEASE (pTexRel);
	RELEASE (pInSurf);
//	RELEASE (pVertexDeclaration);
//	RELEASE (pVB);
	RELEASE (pD3DDev);
	RELEASE (pD3D);
}

void DX9gpgpu::LogInfo()
{
#ifdef INFO_LOG
	D3DADAPTER_IDENTIFIER9 identifier;
	pD3D->GetAdapterIdentifier( D3DADAPTER_DEFAULT, 0, &identifier );
	_snprintf_s( buf, sizeof(buf)/sizeof(TCHAR), MAXSTRSIZE, " %s ( %d x %d ) \n\n", identifier.Description, D3DCaps.MaxTextureWidth, D3DCaps.MaxTextureHeight );
	OutputNote( *buf );
#endif
}

void DX9gpgpu::RenderTest( int width, int height ) // drawing function 
{
	int t=0;

	UINT totalPasses; // Used to store the total amount of passes we need for our effect.
	D3DXHANDLE hTexcoord2;
	LPDIRECT3DSURFACE9 pSwapSurf, pPongSurf;
	LPDIRECT3DTEXTURE9 pSwapTex, pPongTex;
//	int height=1;

	// Correctly map from texels->pixels, offset the coordinates by -0.5f	
	float fWidth = static_cast< float >( width ) - 0.5f;
    float fHeight = static_cast< float >( height ) - 0.5f;    

    CUVERTEX myVertices[] = // Define texcoords: x, y, z, w position points +  2 texture coord
	{ 
		// Disable rectangle easy conditions in shader
		{ -0.5f,   -0.5f,  0.0f, 1.0f, 0.0f, 0.0f},
        { fWidth,  -0.5f,  0.0f, 1.0f, 1.0f, 0.0f},
        { -0.5f,  fHeight, 0.0f, 1.0f, 0.0f, 1.0f},
        { fWidth, fHeight, 0.0f, 1.0f, 1.0f, 1.0f}
    };    

//	if(FAILED(pD3DDev->CreateVertexBuffer(sizeof(myVertices), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &pVB, NULL)))
//		Error(*"Error, couldn't create vertex buffer");

    LPD3DXBUFFER pEB = NULL;
	if (FAILED(D3DXCreateEffectFromFile(pD3DDev, "gpgpugs02_DX9.fx", NULL, NULL, 0, NULL, &pEffect, &pEB)))
    {		
		if (pEB==NULL)
			Error(*"Error, couldn't create effect from file gpgpugs02_DX9.fx");
		else 
		{
			LPVOID compileErrors = pEB->GetBufferPointer();
			Error(*(const char*)compileErrors);
		}
	}
    RELEASE(pEB);

	if (FAILED(D3DXCreateTexture (pD3DDev, width, height, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pPongTex))) // Texture as a rendering target formation D3DFMT_R32F
		Error(*"Error, couldn't create pong texture");
//	ClearTexture(pPongTex,width,height);
	pPongTex->GetSurfaceLevel (0, &pPongSurf);

    pD3DDev->Clear (0L, NULL, D3DCLEAR_TARGET, D3DCOLOR_RGBA (0xff, 0xff, 0xff, 0xff), 1.0f, 0L); // Clear screen

    pD3DDev->BeginScene (); // Tell Direct 3D we are now going to start drawing.
	{
		hTexcoord2 = pEffect->GetParameterBySemantic(NULL, "COLOR1");
//		if (FAILED(pEffect->SetVector (hTexcoord2,&(D3DXVECTOR4 ((float)(width-3)/width, (float)(width-2)/width, (float)(width-2.5)/width, (float)((width-1.5)/width)))))) // Send variable to hlsl - last 4 x coordinates (second ones with .5 correction to condition)
		if (FAILED(pEffect->SetVector (hTexcoord2,&(D3DXVECTOR4 ((float)width, (float)height, (float)1, (float)1))))) // Send variable to hlsl - pivot coordinate
			Error(*"Error, couldn't SetVector");

		if (FAILED(pEffect->SetTexture ("mytex", pInTex))) // Set in texture
			Error(*"Error, couldn't set in texture");

		if (FAILED(pEffect->SetTechnique ("poctyTest")))
			Error(*"Error, couldn't SetTechnique");
		if (FAILED(pD3DDev->SetRenderTarget (0, pPongSurf))) // Modification of rendering target
			Error(*"Error, couldn't SetRenderTarget");

	    pEffect->Begin (&totalPasses, 0);
		for(UINT p=0; p<totalPasses; p++)
		{         
			pEffect->BeginPass(p);
			{						
				pD3DDev->SetFVF(D3DFVF_CUSTOMVERTEX); // Set the vertex stream declaration.
				pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, myVertices, sizeof(CUVERTEX));
			}
			pEffect->EndPass();
		}
		pEffect->End();  // End of the effect.

		// Ping-pong (swap textures)
		pSwapSurf = pInSurf;
		pInSurf = pPongSurf;
		pPongSurf = pSwapSurf;
		pSwapTex = pInTex;
		pInTex = pPongTex;
		pPongTex = pSwapTex;
	}
	pD3DDev->EndScene (); // We are done drawing for this scene.
	pD3DDev->Present (NULL, NULL, NULL, NULL); // Swap the old frame with the new one.

	RELEASE(pPongTex);
}


void DX9gpgpu::RenderMulMat(int width, int height, int mul) // drawing function 
{
	UINT totalPasses; // Used to store the total amount of passes we need for our effect.
	D3DXHANDLE hTexcoord2;
	LPDIRECT3DSURFACE9 pSwapSurf, pPongSurf;
	LPDIRECT3DTEXTURE9 pSwapTex, pPongTex;
//	int height=1;

	// Correctly map from texels->pixels, offset the coordinates by -0.5f	
	float fWidth = static_cast< float >( width ) - 0.5f;
    float fHeight = static_cast< float >( height ) - 0.5f;    

    CUVERTEX myVertices[] = // Define texcoords: x, y, z, w position points +  2 texture coord
	{ 
		{ -0.5f,   -0.5f,  0.0f, 1.0f, 0.0f, 0.0f},
        { fWidth*2,  -0.5f,  0.0f, 1.0f, 2.0f, 0.0f},
        { -0.5f,  fHeight*2, 0.0f, 1.0f, 0.0f, 2.0f},
    };    

//	if(FAILED(pD3DDev->CreateVertexBuffer(sizeof(myVertices), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &pVB, NULL)))
//		Error(*"Error, couldn't create vertex buffer");

    LPD3DXBUFFER pEB = NULL;
	if (FAILED(D3DXCreateEffectFromFile(pD3DDev, "gpgpugs02_DX9.fx", NULL, NULL, 0, NULL, &pEffect, &pEB)))
    {		
		if (pEB==NULL)
			Error(*"Error, couldn't create effect from file gpgpugs02_DX9.fx");
		else 
		{
			LPVOID compileErrors = pEB->GetBufferPointer();
			Error(*(const char*)compileErrors);
		}
	}
    RELEASE(pEB);

	if (FAILED(D3DXCreateTexture (pD3DDev, width, height, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pPongTex))) // Texture as a rendering target formation D3DFMT_R32F
		Error(*"Error, couldn't create pong texture");
//	ClearTexture(pPongTex,width,height);
	pPongTex->GetSurfaceLevel (0, &pPongSurf);

    pD3DDev->BeginScene (); // Tell Direct 3D we are now going to start drawing.
	{
		hTexcoord2 = pEffect->GetParameterBySemantic(NULL, "COLOR1");
		if (FAILED(pEffect->SetVector (hTexcoord2,&(D3DXVECTOR4 ((float)width, (float)height, (float)1, (float)(mul)))))) // Send variable to hlsl
			Error(*"Error, couldn't SetVector");

		if (FAILED(pEffect->SetTexture ("mytex", pInTex))) // Set in texture
			Error(*"Error, couldn't set in texture");

		if (FAILED(pEffect->SetTechnique ("poctyMulMat")))
			Error(*"Error, couldn't SetTechnique");
		if (FAILED(pD3DDev->SetRenderTarget (0, pPongSurf))) // Modification of rendering target
			Error(*"Error, couldn't SetRenderTarget");

	    pEffect->Begin (&totalPasses, 0);
		for(UINT p=0; p<totalPasses; p++)
		{         
			pEffect->BeginPass(p);
			{						
				pD3DDev->SetFVF(D3DFVF_CUSTOMVERTEX); // Set the vertex stream declaration.
				pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, myVertices, sizeof(CUVERTEX));
			}
			pEffect->EndPass();
		}
		pEffect->End();  // End of the effect.

		// Ping-pong (swap textures)
		pSwapSurf = pInSurf;
		pInSurf = pPongSurf;
		pPongSurf = pSwapSurf;
		pSwapTex = pInTex;
		pInTex = pPongTex;
		pPongTex = pSwapTex;
	}
	pD3DDev->EndScene (); // We are done drawing for this scene.
	pD3DDev->Present (NULL, NULL, NULL, NULL); // Swap the old frame with the new one.

	RELEASE(pPongTex);
}
