int computeStyle=0; // Which hardware/algorithm use for computing
int computeIter=1; // Number of iterations

#include "gpgpu02.h"
#include "directx9_gpgpu.h"
#include "directx10_gpgpu.h"
#include "cpu_gpgpu.h"

TCHAR buf[MAXSTRSIZE+1];

char fileNameIn[MAXSTRSIZE+1];
char fileNameOut[MAXSTRSIZE+1];


void Close() // Dispose all before quit
{
}

void Error(const char &text, bool end, int code) // Error box and quit.
{
	_snprintf_s(buf, sizeof(buf)/sizeof(TCHAR), MAXSTRSIZE, &text, code);

	MessageBox(NULL, buf, "Error!", MB_OK);

	// Send string to debugger 
	strncat_s(buf, MAXSTRSIZE+1, " (GS DEBUG)\n", __min( 12, MAXSTRSIZE-strlen(buf)) );
	OutputDebugStringA(buf);

	if (!end)
		return;
	Close();
	exit(-1);
}

void Help(void)
{
	MessageBox(NULL, "Usage: gpgpugs02 [size] [iter] [style] [fixed]\n\n   size = height of created random matrix (it overwrite input.txt!)\n   iter = number of iterations\n   style = how to compute. Possibilities are CPU, CPU2 (=double precision), DX9, DX10, CUDA, FSTR, NONE, PRECCOMP (precision compare).\n U can use list of them.\n\n   If no parameters presents, input.txt will be taken as input matrix and \ncomputation will be done on all three APIs.\n\n   Parameter FIXED means that input_size.txt will be parsed.\n\n   Results are in output.txt", "Help!", MB_OK);
	Close();
	exit(1);
}

int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR pszArgs , int nWinMode) // WinMain function
{
	time_t timeStart, timeEnd, timeRet, timeTmp=0; // Compute time	
	LPWSTR * szArglist;
	int nArgs, i;
	int genSize = defaultGenerate;
	bool doNotCompute = false;
	bool sized = false;
	bool fixedInput = false;
	bool fixedOutput = false;
	int width, height;

	szArglist = CommandLineToArgvW(GetCommandLineW(), &nArgs);
	if (szArglist!=NULL && nArgs>1)
	{
		for( i=1; i<nArgs; i++) 
		{
			if ((_wcsicmp( szArglist[i],L"--help")==0 ) || (_wcsicmp( szArglist[i],L"-h")==0 ) || ( _wcsicmp (szArglist[i],L"help")==0 ) || (_wcsicmp( szArglist[i],L"/h")==0 ) || (_wcsicmp( szArglist[i],L"/help")==0 ))
				Help();

			if ( !sized )
			{
				if (_wtoi( szArglist[i] )>0 ) 
				{
					genSize=_wtoi( szArglist[i] );
					sized = true;
				}
			} else
				if (_wtoi( szArglist[i] )>0 ) 
					computeIter=_wtoi( szArglist[i] );
					
			// bitmask 1 - CPU, 2 - DX9, 4 - DX10, 8 - FSTR, 16 - CUDA
			if ((_wcsicmp( szArglist[i],L"CPU")==0 ))
				computeStyle=computeStyle | 1;
			else if ((_wcsicmp ( szArglist[i],L"DX9")==0))
				computeStyle=computeStyle | 2;
			else if ((_wcsicmp ( szArglist[i],L"DX10")==0))
				computeStyle=computeStyle | 4;
			else if ((_wcsicmp ( szArglist[i],L"FSTR")==0) || (_wcsicmp ( szArglist[i],L"FIRESTREAM")==0))
				computeStyle=computeStyle | 8;
			else if ((_wcsicmp ( szArglist[i],L"CUDA")==0))
				computeStyle=computeStyle | 16;
			else if ((_wcsicmp ( szArglist[i],L"NONE")==0))
				doNotCompute = true;
			else if ((_wcsicmp ( szArglist[i],L"PRECCOMP")==0))
				computeStyle=computeStyle | 32;
			else if ((_wcsicmp ( szArglist[i],L"CPU2")==0))
				computeStyle=computeStyle | 128;
			else if ((_wcsicmp ( szArglist[i],L"MULMAT")==0))
				computeStyle=computeStyle | 256;
			else if ((_wcsicmp ( szArglist[i],L"FIXED")==0))
			{
				fixedInput = true;
				fixedOutput = true;
			}
			else if ((_wcsicmp ( szArglist[i],L"FIXEDOUT")==0))
				fixedOutput = true;
			else if ((_wcsicmp ( szArglist[i],L"FIXEDIN")==0))
				fixedInput = true;
		}
		LocalFree(szArglist); // Free memory allocated for CommandLineToArgvW arguments.
	}

	_snprintf_s( fileNameIn, sizeof(fileNameIn)/sizeof(TCHAR), MAXSTRSIZE, "input.txt" );
	_snprintf_s( fileNameOut, sizeof(fileNameOut)/sizeof(TCHAR), MAXSTRSIZE, "output.txt" );
	if ( fixedInput )
	{
		if ( genSize ) // Get matrix from input_XXX.txt
		{
			_snprintf_s( fileNameIn, sizeof(fileNameIn)/sizeof(TCHAR), MAXSTRSIZE, "input_%d.txt", genSize );
		}
	} else
		if ( genSize ) // Generate input
			GenerateRandomInput(genSize );
	if ( fixedOutput )
	{
		if ( computeIter>1 )
			_snprintf_s( fileNameOut, sizeof(fileNameOut)/sizeof(TCHAR), MAXSTRSIZE, "output_%d_%d.txt", genSize, computeIter );
		else
			_snprintf_s( fileNameOut, sizeof(fileNameOut)/sizeof(TCHAR), MAXSTRSIZE, "output_%d.txt", genSize );
	}

	if ( doNotCompute )
	{
		Close();
		exit(0);
	}

	if ( computeStyle==0 )
		computeStyle=defaultComputeStyle;

	// Create output.txt
	OutputNote(NULL);
	OutputNoteDefault(NULL);

	FILE * fp;
	if (fopen_s (&fp, fileNameIn, "r")!=0)
		Help();

	fscanf_s(fp,"%d %d\n",&width,&height);

#ifdef INFO_LOG
	_snprintf_s( buf, sizeof(buf)/sizeof(TCHAR), MAXSTRSIZE, "size: %d x %d\niterations: %d\nipnut div: %d\n\n", width, height, computeIter, INPUT_DIV );
	OutputNote( *buf );
#endif
	fclose(fp);

	if ( computeStyle == 32 && width<4) // Only preccomp don't need input
		Error(*"PRECCOMP need minimum 4x1 matrix size");

	HWND hWnd; // Handle to the window.
	WindowCreate(hThisInst, &hWnd);
	ShowWindow(hWnd, SW_SHOW);    // Show the window.

	// Decide how to compute
//	Priority(1);
	if (computeStyle & 2) // DX9
	{

		DX9gpgpu DX9;

		OutputNote(*"DX9:\n");

		DX9.InitDirect3D( hWnd );
		DX9.LogInfo();
		DX9.InputData( fileNameIn );
		UpdateWindow(hWnd); // Update its display.

		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			DX9.Render( width, height );
			timeEnd=GetTickCount();

			if ( !i || i==(computeIter-1)) // Log first and last only
			{
				DX9.OutputData(fileNameOut);
			}
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			DX9.ReloadData();
		}
//		WindowThread(); // Wait for escape
	} 
	if (computeStyle & 4) // dx10
	{
		OutputNote(*"DX10:\n");
#ifdef COMPILE_DX10
		DX10gpgpu DX10;

		DX10.InitDirect3D(hWnd);
		DX10.LogInfo();
		DX10.InputData( fileNameIn );

		UpdateWindow(hWnd); // Update its display.

		for ( int i=0; i<computeIter; i++ )
		{
	        timeStart=GetTickCount();
			DX10.Render();
			timeEnd=GetTickCount();

			if ( !i || i==(computeIter-1)) // Log first and last only
				DX10.OutputData(fileNameOut); // It outputs to the file

			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			DX10.ReloadData();
		}
#else
		OutputNote(*"  not compiled for use.\n\n\n");
#endif
	}
	if (computeStyle & 8) // firestream
	{
		OutputNote(*"FireStream:\n");
		OutputNote(*"  not supported (yet).\n\n\n");
	}
	if (computeStyle & 16) // cuda
	{
		OutputNote(*"CUDA:\n");
		OutputNote(*"  not supported (yet).\n\n\n");
	}
	if (computeStyle & 1) // cpu single precision
	{
		CPUgpgpu<float> CPU;

		OutputNote(*"CPU:\n");

		CPU.LogInfo();
		CPU.InputData( fileNameIn );
		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			timeRet=CPU.Compute();
			timeEnd=GetTickCount();
//			timeTmp+=timeRet;
//			OutputNoteInt( (int)(timeRet) );
//			OutputNoteNL();

			if ( !i || i==(computeIter-1)) // Log first and last only
				CPU.OutputData(fileNameOut);
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			CPU.ReloadData();
		}
//		OutputNoteInt( (int)(timeTmp) );
	}
	if (computeStyle & 128) // cpu double precision
	{
		CPUgpgpu<double> CPU;

		OutputNote(*"CPU (double prec):\n");

		CPU.LogInfo();
		CPU.InputData( fileNameIn );
		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			timeRet=CPU.Compute();
			timeEnd=GetTickCount();
//			timeTmp+=timeRet;
//			OutputNoteInt( (int)(timeRet) );
//			OutputNoteNL();

			if ( !i || i==(computeIter-1)) // Log first and last only
				CPU.OutputData(fileNameOut);
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			CPU.ReloadData();
		}
//		OutputNoteInt( (int)(timeTmp/1000) );
	}
	if (computeStyle & 32) // Precision compare (CPU vs DX9)
	{
		double a=30;
		double b=65;

		{
			DX9gpgpu DX9;
			DX9.InitDirect3D( hWnd );
			DX9.LogInfo();
		}

		for ( int i=0; i<6; i++)
		{
			_snprintf_s( buf, sizeof(buf)/sizeof(TCHAR), MAXSTRSIZE, "=====%.4f x %.4f===== ( DOUBLE / FLOAT / DX9 ) \n", a, b );
			OutputNote( *buf );

			{
				CPUgpgpu<double> CPU;

				CPU.InputData( fileNameIn, true ); // Hack - must create input matrix even if not used input data
				CPU.ComputeTest( a, b );

				CPU.OutputData( fileNameOut, LOG_TYPE_PRECCOMP );
			}
			{
				CPUgpgpu<float> CPU;

				CPU.InputData( fileNameIn, true ); // Hack - must create input matrix even if not used input data
				CPU.ComputeTest( a, b );

				CPU.OutputData( fileNameOut, LOG_TYPE_PRECCOMP );
			}
			{
				DX9gpgpu DX9;

				DX9.InitDirect3D( hWnd );
				DX9.InputDataTest( fileNameIn, a, b ); // Hack - must create input matrix even if not used input data

				UpdateWindow(hWnd);	
				DX9.RenderTest( width, height);

				DX9.OutputData( fileNameOut, LOG_TYPE_PRECCOMP );
			}
			OutputNote(*"\n");
			a = a/10;
			b = b/10;
		}
	}
/*
	if (computeStyle & 64) // DX9 texture multiplication
	{

		DX9gpgpu DX9;

		OutputNote(*"DX9 tex multiply:\n");

		DX9.InitDirect3D( hWnd );
		DX9.InputData( fileNameIn );
		UpdateWindow(hWnd); // Update its display.

		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			DX9.RenderMulti( width, height );
			timeEnd=GetTickCount();

			if ( !i || i==(computeIter-1)) // Log first and last only
			{
				DX9.OutputData(fileNameOut);
			}
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			DX9.ReloadData();
		}
//		WindowThread(); // Wait for escape
	} 
*/
	if (computeStyle & 256) // nasobeni matic
	{
		CPUgpgpu<float> CPU;

		OutputNote(*"CPU mul mat:\n");

		CPU.LogInfo();
		CPU.InputData( fileNameIn );
		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			CPU.ComputeMulMat( 3 );
			timeEnd=GetTickCount();

			if ( !i || i==(computeIter-1)) // Log first and last only
				CPU.OutputData(fileNameOut);
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			CPU.ReloadData();
		}
		DX9gpgpu DX9;

		OutputNote(*"DX9 mul mat:\n");

		DX9.InitDirect3D( hWnd );
		DX9.LogInfo();
		DX9.InputData( fileNameIn );
		UpdateWindow(hWnd); // Update its display.

		for ( int i=0; i<computeIter; i++ )
		{
			timeStart=GetTickCount();
			DX9.RenderMulMat( width, height, 3 );
			timeEnd=GetTickCount();

			if ( !i || i==(computeIter-1)) // Log first and last only
				DX9.OutputData(fileNameOut);
			OutputTime( timeStart, timeEnd, i==(computeIter-1));
			DX9.ReloadData();
		}
	}
//	Priority(2);
	// End of all computing
	UnregisterClass("gpgpugsClass", hThisInst); // Here we unregister the window class with the OS.
	DestroyWindow(hWnd); // After that MessageBox stop working

	return (0);
}

void WindowCreate (HINSTANCE hThisInst, HWND* hWnd) // Compilation of window
{
	WNDCLASSEX wcl; // Window class object.

	wcl.cbSize = sizeof (WNDCLASSEX); // size of the WNDCLASSEX structure.
	wcl.style = 0; // style of the window.
	wcl.lpfnWndProc = WinProc; // Address to the windows procedure.
	wcl.cbClsExtra = 0; // Extra class information.
	wcl.cbWndExtra = 0; // Extra window information.
	wcl.hInstance = hThisInst; // Handle of application Instance.
	wcl.hIcon = LoadIcon (NULL, IDI_APPLICATION); // Window Icon.
	wcl.hCursor = LoadCursor (NULL, IDC_ARROW); // Load mouse cursor.
	wcl.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH); // Background color.
	wcl.lpszMenuName = NULL; // Menu.
	wcl.lpszClassName = "gpgpugsClass"; // Name of the window class.
	wcl.hIconSm = LoadIcon (NULL, IDI_WINLOGO); // Window icon.

	if (!RegisterClassEx (&wcl)) // You must register you class with Windows.
		Error(*"Window class can't be registered");

	*hWnd = CreateWindow ( // After your class has been registered then you are ready to create your window.
		"gpgpugsClass", // Window class.
		"gpgpugs", // Window name.
		WS_OVERLAPPEDWINDOW, // Window style.
		0, 0, // X, Y coords.
		64, 64, // Window size.
//		HWND_MESSAGE, 
		HWND_DESKTOP, // Handle to parent window.
		NULL, // Menu.
		hThisInst, // Handle to app instance.
		NULL // Pointer to window.
	);
                            
	if(!hWnd) // Error with creating the window.
		Error(*"Error, couldn't initialize window!?!");
}

LRESULT CALLBACK WinProc (HWND hWnd, UINT msg, WPARAM wP, LPARAM lP) // window procedure 
{
	switch (msg) 
	{
		case WM_KEYUP:
			if(wP==VK_ESCAPE) // If the user presses the escape key then exit the application.
				PostQuitMessage(0);
			break;
		case WM_DESTROY:
		case WM_CLOSE:
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc (hWnd, msg, wP, lP);
	} 
	return 0;
}

void OutputTime( time_t timeStart, time_t timeEnd, bool print_and_null )
{
	static time_t timeSum=0;
	timeSum += ( timeEnd-timeStart );
	if ( print_and_null )
	{
		OutputNote(*"time: ");
		OutputNoteInt( (int)(timeSum/1000) );
		OutputNote(*"s\n\n\n");	
		timeSum=0;
	}
}

void GenerateRandomInput(int height) // Random array
{
	FILE* fp;
    const int RANDMAX = 1000;
	int colnr = 0;
	int width = height+1;

	srand( (unsigned)GetTickCount());

	if (fopen_s (&fp, fileNameIn, "w")!=0)
		Error(*"Can't generate input.txt");

	fprintf(fp,"%d %d\n",width,height);

	for (int j=0; j<height; j++) {
		for (int i=0; i<width; i++) {
#ifndef GENERATE_NON_RANDOM
			// Random input
			fprintf(fp,"%d ",(int)((rand()*RANDMAX/RAND_MAX)+1)); // +1 cause I don't want zero on diagonal :)
#else
			// All variables = 1
			int nr=0;
			if (i==height) // last column
			{
				fprintf(fp,"%d ",colnr);
				colnr=0;
			} else
			{
				if (i<j)
					nr=i+j;
				else
					nr=i-j+1;
/*				if (i<=j) // below diagonal
					nr=1; // +j*2;
				else
					nr=(i-j+1);
*/				fprintf(fp,"%d ",nr);
				colnr+=nr*(i+1);
			}
#endif
		}
		fprintf(fp,"\n");
	}

	fclose(fp);
}

void OutputNoteDefault(const char &note)
{
	FILE* fp;
	if (note==NULL)
		fopen_s( &fp, "output.txt", "w" );
	else
	{
		fopen_s( &fp, "output.txt", "a" );
		fprintf( fp, "%s", &note );
	}
	fclose(fp);
}

void OutputNote(const char &note)
{
	FILE* fp;
	if (note==NULL)
		fopen_s( &fp, fileNameOut, "w" );
	else
	{
		fopen_s( &fp, fileNameOut, "a" );
		fprintf( fp, "%s", &note );
	}
	fclose(fp);
}

void OutputNoteNL()
{
	FILE* fp;
	fopen_s (&fp, fileNameOut, "a");
	fprintf(fp,"\n");
	fclose(fp);
}

void OutputNoteInt(DWORD64 note)
{
	FILE* fp;
	fopen_s (&fp, fileNameOut, "a");
	fprintf(fp,"%d",note);
	fclose(fp);
}

int WindowThread()
{
	MSG msg; // Message object.
	ZeroMemory(&msg, sizeof (msg)); // It is always a good idea to clear out memory in object although not necessary.

	while(true) // Wait for escape
	{
		if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
		{
			if(msg.message==WM_QUIT) // quit message received
				break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else
			Sleep(100);
	}
    return (int) msg.wParam;
}

void Priority(short int i) //0=realtime; 1=high; 2=normal
{
	switch ( i ) 
	{
		case 0:
			SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
			break;
		case 1:
			SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
			break;
		default:
		case 2:
			SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
			break;
	}
}