//// states
BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

DepthStencilState DisableDepth
{
    DepthEnable = FALSE;
    DepthWriteMask = ZERO;
};

RasterizerState DisableCulling
{
	FillMode = SOLID;
	CullMode = NONE;
	DepthClipEnable = FALSE;
};

//// textures
Texture2D mytexTemp;
Texture2D mytex;

//// samplers
SamplerState samNothing
{
	Filter = MIN_MAG_MIP_POINT;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

//// inputs, outputs
uniform int pivot;  // Pivot passed from DX
uniform int width;
uniform int height;

struct VS_OUTPUT // This structure will define how we are sending the data to the pixel shader.
{ 
    float4 Pos: SV_POSITION;
	float2 Tex: TEXCOORD0;
};

//// vertex shader
VS_OUTPUT VSCopy( float4 Pos : POSITION , float2 Tex : TEXCOORD0)
{
	VS_OUTPUT Out;
	Out.Pos = Pos;
	Out.Tex = Tex;
	return Out;
}

//// pixel shader
/*
float4 PSCopy( VS_OUTPUT In ) : SV_TARGET
{
	float4 Out;
	Out = mytex.Sample( samNothing, float2(In.Tex.x, In.Tex.y) );
	return Out;
}
*/
float4 PSSubMat ( VS_OUTPUT In ) : SV_TARGET // inMatrix[j+i*width] = inMatrix[j+i*width] - tmpMatrix[t+i*width] * inMatrix[j+t*width];
{
	float4 Out;
	float inMatA, tmpMat, inMatB, inMatABranch;
	
	//------- Old DX9 access to texture
	float pivotY, pivotX, pivotYplus, pivotXminus;

	pivotX = (pivot+.5)/width; //texel start in half pixel
	pivotY = (pivot+.5)/height;
	pivotXminus = (pivot-0.5)/width;
	pivotYplus = (pivot+0.5)/height;
	
	inMatB = mytex.Sample( samNothing, float2(In.Tex.x, pivotY) ); // inMatrix[j+t*width]
	tmpMat = mytexTemp.Sample( samNothing, float2(0.5, In.Tex.y) );	
	//---------------

	// Original DX10 access to texture
/*	inMatB = mytex.Load( int3(In.Tex.x*width, pivot, 0) );
	tmpMat = mytexTemp.Sample( samNothing, In.Tex.xy ); // tmpMatrix[t+i*width]
	//---------------
*/	
	inMatA = mytex.Sample( samNothing, In.Tex ); // inMatrix[j+i*width]  
	inMatABranch = inMatA-tmpMat*inMatB;	
	
    // for (int i=(t+1); i<height; i++) for (int j=t; j<width; j++)
//	[branch] // insert code that let the shader decide at runtime which side should taken
//	if (In.Tex.y>pivotYplus && In.Tex.x>pivotXminus) // submat
	if (In.Tex.y*height>(pivot+1) && In.Tex.x*width>pivot) // submat
	{
		inMatA = inMatABranch;
//		inMatA = inMatA-tmpMat*inMatB;
//		inMatA = mytex.Sample( samNothing, In.Tex )+0.1;
	} 
//	else // original
//		inMatA = mytex.Sample( samNothing, In.Tex );
//		inMatA = 255;
		
//	inMatA = tmpMat + 0.1; // DEBUG

	Out = ( 0, 0, 0, 0 );
	Out.r = inMatA; // inMatrix[j+i*width]

	return Out;
}

float4 PSRow ( VS_OUTPUT In ) : SV_TARGET // tmpMatrix[t+i*width] = inMatrix[t+i*width] / pivot;
{
	float4 Out;
	float pVal,inMat;
	
	// Same access as DX9
	float pivotX,pivotY;
	
	pivotX = (pivot+.5)/width; //texel start in half pixel
	pivotY = (pivot+.5)/height;

	inMat = mytex.Sample( samNothing, float2(pivotX, In.Tex.y) ); // inMatrix[t+i*width]
	pVal = mytex.Sample( samNothing, float2(pivotX, pivotY) ); // inMatrix[t+t*width]

	// New DX10 texture access
/*	inMat = mytex.Load( int3(pivot, In.Tex.y*height, 0) ); // inMatrix[t+i*width]
	pVal = mytex.Load( int3(pivot, pivot, 0) ); // inMatrix[t+t*width]
*/
	Out = ( 0, 0, 0, 0 );

	// for (int i=(t+1); i<height; i++)
//	if (In.Tex0.y>pivotPos.a) // row
		Out.r = inMat / pVal; // tmpMatrix[t+i*width]
//	else // otherwise doesn't matter
//		Out.Color.r = tex2D(mySamp, In.Tex0); // inMatrix[j+i*width]

//	Out.r = In.Tex.y; // DEBUG

	return Out;
}

//// Technique
technique10 poctySubMat
{
    pass P0
    {
		// Set VS, GS, and PS
        SetVertexShader( CompileShader( vs_4_0, VSCopy() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSSubMat() ) );
//		SetPixelShader( CompileShader( ps_4_0, PSCopy() ) );
		
		// Set our render states
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );
        SetRasterizerState( DisableCulling );
    }
}

technique10 poctyRow
{
    pass P0
    {
		// Set VS, GS, and PS
        SetVertexShader( CompileShader( vs_4_0, VSCopy() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSRow() ) );
//		SetPixelShader( CompileShader( ps_4_0, PSCopy() ) );
		
		// Set our render states
		SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
		SetDepthStencilState( DisableDepth, 0 );
		SetRasterizerState( DisableCulling );
    }
}